#define PROGNAME "ED_jacobi_nocorr"
#define VERSION "0.3"
#define DATE  "14.06.2007"
#define AUTHOR "Nils Bluemer"

/* performs Jacobi sweep (for determination of eigenvalues)
   NEW (0.2): only output of energies at verbosity=0 
   NEW (0.3): uses pointer_utils.c, comments added 
   VERSION nocorr: computation of correlation function removed */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "pointer_utils.c"

void error(char error_text[])
/* standard error handler */
{
  fprintf(stderr,"\n %s run-time error\n", PROGNAME);
    fprintf(stderr,"--%s--\n",error_text);
    fprintf(stderr,"for general help use option -h\n");
    fprintf(stderr,"...now exiting to system...\n");
    exit(1);
}

void printhelp ()
{
 printf("**********************************************************\n");
 printf("%s: eigenvalue utility\n",PROGNAME);
 printf("Version: %s, %s by %s\n",VERSION,DATE,AUTHOR);
 printf("Input: symmetric matrix on stdin, preceded by dimension,\n");
 printf("e.g.: 3\n      1 2 3\n      2 4 5\n      3 5 6\n");  
 printf("options: -d# digits on output\n");
 printf("         -v# verbosity (default:1)\n");
 printf("         -c compute NN correlation for ground state\n");
 printf("         -h this help\n");
}

void print_matrix(double **a, int size, int digits)
{
  int i,j;
  char buffer [20];

  sprintf(buffer,"%%%d.%dlf",digits+4,digits);
/*   printf("%s\n",buffer); */

  for (i=1;i<=size;i++){
    for (j=1;j<=size;j++)
      printf(buffer,a[i][j]);
    printf("\n");
  }
}

void print_comp_matrix(double **a, double *d, int size, int digits)
/* prints matrix after some Jacobi rotations (reconstructed from a and d) */
{
  int i,j;
  char buffer [20];

  sprintf(buffer,"%%%d.%dlf",digits+4,digits);
/*   printf("%s\n",buffer); */

  for (i=1;i<=size;i++){
    for (j=1;j<=size;j++)
      if (i<j)
	printf(buffer,a[i][j]);
      else 
	if (i>j)
	  printf(buffer,a[j][i]);
	else 
	  printf(buffer,d[i]);
    printf("\n");
  }
}


void print_vector(double *v, int size, int digits)
{
  int i;
  char buffer [20];

  sprintf(buffer,"%%%d.%dlf ",digits+4,digits);
  for (i=1;i<=size;i++)
    printf(buffer,v[i]);
  printf("\n");
}


#define ROTATE(a,i,j,k,l) g=a[i][j];h=a[k][l];a[i][j]=g-s*(h+g*tau);\
	a[k][l]=h+s*(g-h*tau);

void jacobi(double **a, int n, double d[], double **v, int *nrot, int *sweeps, int verbosity, int digits) 
/* standard routine for diagonalization, destroys input matrix a on output
   d: vector of eigenvalues, v: array of eigenvectors, n: dimension,
   nrot: number of jacobi rotations, sweeps: number of iterations (sweeps),
   verbosity: determines output, digits: determines precision on output */
{
  int j,iq,ip,i;
  double tresh,theta,tau,t,sm,s,h,g,c,*b,*z;

  b=dvector(1,n);
  z=dvector(1,n);
  for (ip=1;ip<=n;ip++) {
    for (iq=1;iq<=n;iq++) v[ip][iq]=0.0;
    v[ip][ip]=1.0;
  }
  for (ip=1;ip<=n;ip++) {
    b[ip]=d[ip]=a[ip][ip];
    z[ip]=0.0;
  }
  *nrot=0;
  for (i=1;i<=50;i++) {
    sm=0.0;
    for (ip=1;ip<=n-1;ip++) {
      for (iq=ip+1;iq<=n;iq++)
	sm += fabs(a[ip][iq]);
    }
    if (sm == 0.0) {
      free_dvector(z,1,n);
      free_dvector(b,1,n);
      return;
    }
    if (i < 4)
      tresh=0.2*sm/(n*n);
    else
      tresh=0.0;
    for (ip=1;ip<=n-1;ip++) {
      for (iq=ip+1;iq<=n;iq++) {
	g=100.0*fabs(a[ip][iq]);
	if (i > 4 && (double)(fabs(d[ip])+g) == (double)fabs(d[ip])
	    && (double)(fabs(d[iq])+g) == (double)fabs(d[iq]))
	  a[ip][iq]=0.0;
	else if (fabs(a[ip][iq]) > tresh) {
	  h=d[iq]-d[ip];
	  if ((double)(fabs(h)+g) == (double)fabs(h))
	    t=(a[ip][iq])/h;
	  else {
	    theta=0.5*h/(a[ip][iq]);
	    t=1.0/(fabs(theta)+sqrt(1.0+theta*theta));
	    if (theta < 0.0) t = -t;
	  }
	  c=1.0/sqrt(1+t*t);
	  s=t*c;
	  tau=s/(1.0+c);
	  h=t*a[ip][iq];
	  z[ip] -= h;
	  z[iq] += h;
	  d[ip] -= h;
	  d[iq] += h;
	  a[ip][iq]=0.0;
	  for (j=1;j<=ip-1;j++) {
	    ROTATE(a,j,ip,j,iq)
	      }
	  for (j=ip+1;j<=iq-1;j++) {
	    ROTATE(a,ip,j,j,iq)
	      }
	  for (j=iq+1;j<=n;j++) {
	    ROTATE(a,ip,j,iq,j)
	      }
	  for (j=1;j<=n;j++) {
	    ROTATE(v,j,ip,j,iq)
	      }
	  ++(*nrot);
	  if (verbosity>=4){
	    printf("Matrix A after %d Jacobi sweeps (%d rotations):\n",*sweeps,*nrot);
	    print_comp_matrix(a,d,n,digits);
	  }
	}
      }
    }
    for (ip=1;ip<=n;ip++) {
      b[ip] += z[ip];
      d[ip]=b[ip];
      z[ip]=0.0;
    }
    (*sweeps)++;
    if (verbosity>=2){
      printf("Matrix A after %d Jacobi sweeps (%d rotations):\n",*sweeps,*nrot);
      print_comp_matrix(a,d,n,digits);
    }
    if (verbosity>=3){
      printf("Matrix v after %d Jacobi sweeps (%d rotations):\n",*sweeps,*nrot);
      print_matrix(v,n,digits);
    }
  }
  error("Too many iterations in routine jacobi");
}

void eigsrt(double d[], double **v, int n)
/* sorts eigenvalues and eigenvectors by eigenvalues (largest first) */
{
	int k,j,i;
	double p;

	for (i=1;i<n;i++) {
		p=d[k=i];
		for (j=i+1;j<=n;j++)
			if (d[j] >= p) p=d[k=j];
		if (k != i) {
			d[k]=d[i];
			d[i]=p;
			for (j=1;j<=n;j++) {
				p=v[j][i];
				v[j][i]=v[j][k];
				v[j][k]=p;
			}
		}
	}
}


void read_matrix(double **a, int size)
{
  int i,j;
  double dummy;

  for (i=1;i<=size;i++)
    for (j=1;j<=size;j++){
      scanf("%lf", &dummy);
      a[i][j]=dummy;
    }
}

int main (int argc, char *argv[])
{
  char c;
  int i;
  int digits, size, nrot, sweeps, verbosity, corr;
  double **a, **v, *d, *a_diag, x, prob;
  

  digits=5;
  verbosity=2;
  corr=0;

  while (--argc > 0 && (*++argv)[0] == '-')
    while (c= *++argv[0])
	   switch (c) {
	   case 'd':
 	     sscanf(++argv[0],"%d\n",&digits); 
             break;
	   case 'c':
 	     corr=1; 
             break;
	   case 'v':
 	     sscanf(++argv[0],"%d\n",&verbosity); 
             break;
	   case 'h':
	     printhelp();
	     exit(0);
/*  	   default:  */
/*  	     error("No valid choice");  */
	   }

  scanf("%d",&size);
  a=dmatrix(1,size,1,size);
  v=dmatrix(1,size,1,size);
  d=dvector(1,size);

  read_matrix(a,size);

  if (verbosity>0){
    printf("Original matrix A:\n");
    print_matrix(a,size,digits);
    printf("Eigenvalues:\n");
  }

  if (corr>0){ /* IMPLEMENT! */
  }
    

  nrot=0; sweeps=0;

  jacobi(a,size,d,v,&nrot,&sweeps,verbosity,digits);
    
  eigsrt(d,v,size);

  print_vector(d,size,digits);

  if (verbosity>=1){
    printf("final Eigenvectors:\n");
    print_matrix(v,size,digits);
  }

  if (corr>0){/* IMPLEMENT! */
    printf("# N * nearest-neighbor correlation (E_Ising): IMPLEMENT!\n", x);
  }

  free_dmatrix(a,1,size,1,size);
  free_dmatrix(v,1,size,1,size);
  free_dvector(d,1,size);
}

